from importlib import resources
from typing import Iterable

import jieba
import jieba.posseg as pseg

from ..lang import to_simplified

_bible_tokens_loaded: bool = False


def word_tokenize(sentence: str) -> Iterable[str]:
    global _bible_tokens_loaded
    if not _bible_tokens_loaded:
        for token_file in ["classics.txt", "names.txt"]:
            tokens = resources.open_text(__package__, token_file)
            jieba.load_userdict(tokens)
        _bible_tokens_loaded = True

    sentence_s = to_simplified(sentence)
    assert len(sentence) == len(sentence_s)
    total = 0
    for tk in pseg.cut(sentence_s):
        tk.word = sentence[total : total + len(tk.word)]
        total += len(tk.word)
        if tk.flag != "x":
            yield tk.word



import pickle
from importlib import resources
from itertools import groupby, product
from typing import List

import numpy as np

from .resources.db import ccn_embeddings
from .search import BibleSearchStrategy, Match
from .word_tokenizer import word_tokenize


class ConceptNetStrategy(BibleSearchStrategy):
    def __init__(self):
        """
        Initialize search required data, it can be saved and re-loaded to skip
        initialization every time. It has 3 steps:
        1. Word tokenize the whole Bible (turn sentences into word pieces)
        2. Get word vectors of words exist in the Bible
        3. Tokenize each word in the Bible (use a unique number to repsent each word)
           for performance reason
        """
        from .resources import bible

        print("Word tokenizing verses...")
        dots = r"[•‧．・\-]"
        verses = bible.text.str.replace(dots, "", regex=True)
        word_tokenized_verses = verses.transform(lambda v: list(word_tokenize(v)))

        print("Getting word vectors...")
        self._all_words = np.unique(
            [word for verse in word_tokenized_verses for word in verse]
        )
        self._words_vec = ccn_embeddings.get_word_vectors(self._all_words)

        print("Tokenizing verses...")
        tokenized_verses = [
            list(set(map(self._tokenize, verse))) for verse in word_tokenized_verses
        ]
        max_length = max(map(len, tokenized_verses))
        self._tokenized_verses = np.array(
            [v + [0] * (max_length - len(v)) for v in tokenized_verses]
        )

    def to_pickle(self, path: str):
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def from_pickle(cls):
        with resources.open_binary("ezra.resources", "conceptnet_strategy.pickle") as f:
            return pickle.load(f)

    def search(
        self, keyword: str, top_k: int, in_range: np.ndarray = None
    ) -> List[Match]:
        similarity = self.compute_similarity(keyword)

        verses_in_range = (
            self._tokenized_verses[in_range]
            if in_range is not None
            else self._tokenized_verses
        )
        # there are two parts of score: cosine similarity and count of good keywords
        all_match_scores = similarity[:, verses_in_range]
        match_kw_idx = all_match_scores.argmax(axis=2).T
        match_tokens = np.take_along_axis(verses_in_range, match_kw_idx, axis=1).T
        kw_verse_scores = np.take_along_axis(similarity, match_tokens, axis=1)
        verse_scores = np.core.records.fromarrays(
            [
                kw_verse_scores.sum(axis=0),
                np.where(kw_verse_scores >= 0.5, 1, 0).sum(axis=0),
            ],
            names="total,good_kw_counts",
        )
        verse_sorted_idx = np.flip(
            np.argsort(verse_scores, order=["good_kw_counts", "total"])
        )
        top_matches = verse_sorted_idx[:top_k]

        def create_match(index: int) -> Match:
            match_kw = map(self._detokenize, match_tokens.T[index])
            kw_scores = list(zip(match_kw, kw_verse_scores.T[index]))
            index_global = in_range[index] if in_range is not None else index
            return Match(index_global, kw_scores)

        return [create_match(i) for i in top_matches]

    def related_keywords(self, keyword: str, top_k: int = 5) -> List[str]:
        similarity = self.compute_similarity(keyword)
        relevant_tk = np.argwhere(similarity >= 0.5)
        tk_groups = []
        for _, g in groupby(relevant_tk, lambda r: r[0]):
            tk_groups.append(list(map(lambda r: r[1], g)))
        tk_combinations = np.array(list(product(*tk_groups)))
        if tk_combinations.size > 0:
            tk_scores = np.take_along_axis(similarity, tk_combinations.T, axis=1)
            total_scores = tk_scores.sum(axis=0)
            tk_idx = np.flip(np.argsort(total_scores))
            top_suggestions = tk_combinations[tk_idx]
            suggestions = [" ".join(s) for s in map(self._detokenize, top_suggestions)]
            return [s for s in suggestions if s != keyword][:top_k]
        else:
            return []

    def compute_similarity(self, keyword: str) -> np.ndarray:
        keyword_tk = np.array(list(word_tokenize(keyword)))
        kw_vec = ccn_embeddings.get_word_vectors(keyword_tk)

        reserved_tokens = np.zeros((keyword_tk.size, 1))
        exact = pairwise_equal(keyword_tk, self._all_words)
        cosine = self._pairwise_cosine_similarity_words_vec(kw_vec)
        concat = np.where(exact, 1, cosine)
        return np.hstack((reserved_tokens, concat))

    # to reduce load time of Sci-Kit Learn, cosine similarity was implemented
    # the result should be the same as `sklearn.metrics.pairwise.cosine_similarity()`
    def _pairwise_cosine_similarity_words_vec(self, kw_vec: np.ndarray) -> np.ndarray:
        if not hasattr(self, "_words_vec_normalized"):
            self._words_vec_normalized = normalize(self._words_vec).T
        return normalize(kw_vec).dot(self._words_vec_normalized)

    # reserve token for padding
    _reserved_token_length = 1

    def _tokenize(self, word: str) -> int:
        index = np.argmax(self._all_words == word)
        return ConceptNetStrategy._reserved_token_length + index

    def _detokenize(self, token: int) -> str:
        index = token - ConceptNetStrategy._reserved_token_length
        return self._all_words[index]


def pairwise_equal(xs: np.array, ys: np.array) -> np.ndarray:
    return np.stack([ys == x for x in xs])


def normalize(X: np.ndarray) -> np.ndarray:
    X = X.astype(float)
    l2_norms = np.einsum("ij,ij->i", X, X)
    np.sqrt(l2_norms, l2_norms)
    l2_norms[l2_norms == 0.0] = 1.0
    X /= l2_norms[:, np.newaxis]
    return X



import atexit
from importlib import resources
from typing import Iterable

import numpy as np
import tables

from ..lang import to_simplified
from . import abbr


class Bible:
    def __init__(self, db: tables.File):
        self._unv = db.root.unv.table
        self._book = db.root.unv.meta.book.meta.table
        self.idx = np.arange(self._unv.nrows)

    def get_record(self, index: int):
        _, (chap, vers), book_index, text = self._unv[index]
        _, book = self._book[book_index]
        return {
            "book": book.decode(),
            "chap": int(chap),
            "vers": int(vers),
        }, text.decode()

    def exact_match(self, keyword: str) -> np.ndarray:
        cond = "&".join("contains(text, %r)" % kw.encode() for kw in keyword.split())
        return self._unv.read_where(cond, field="index")

    def book_idx(self, in_book: str) -> np.ndarray:
        books = np.char.encode(abbr.ranges.get(in_book, [in_book]))
        all_books = self._book.read(field="values")
        book_indices = np.argwhere(np.isin(all_books, books))
        book_col = self._unv.read(field="book")
        return np.argwhere(np.isin(book_col, book_indices))


class ConceptNetEmbeddings:
    def __init__(self, db: tables.File):
        self._emb = db.root.zh.table

    def get_word_vectors(self, words: Iterable[str]) -> np.ndarray:
        vectors = list(map(self.get_word_vector, words))
        return np.vstack(vectors)

    def get_word_vector(self, word: str) -> np.ndarray:
        key = f"/c/zh/{to_simplified(word)}".encode()  # noqa: F841
        records = self._emb.read_where("index == key")
        if len(records) > 0:
            word, vector = records[0]
            return vector
        else:
            return np.zeros(300, "i1")


with resources.path(__package__, "db.h5") as h5_path:
    _db = tables.open_file(h5_path)
    atexit.register(_db.close)
    bible = Bible(_db)
    ccn_embeddings = ConceptNetEmbeddings(_db)



import sys

import pandas as pd
from ezra.resources import bible


def create_zh_table(conceptnet_h5: str, db_file):
    full = pd.read_hdf(conceptnet_h5)
    zh = full.index.str.startswith("/c/zh")
    full[zh].to_hdf(db_file, "zh", format="table")


def create_unv_table(db_file):
    bible.df.to_hdf(db_file, "unv", format="table", data_columns=["book", "text"])


if __name__ == "__main__":
    path = sys.argv[1]
    with pd.HDFStore(path, mode="w") as db_file:
        create_zh_table("data/mini.h5", db_file)
        create_unv_table(db_file)



import json

from flask import Flask, redirect, render_template, request

from .conceptnet_strategy import ConceptNetStrategy
from .lang import to_simplified  # noqa: F401
from .resources import abbr
from .search import BibleSearchEngine, BibleSearchStrategy, Match  # noqa: F401
from .word_tokenizer import word_tokenize  # noqa: F401

strategy = ConceptNetStrategy.from_pickle()
ezra_engine = BibleSearchEngine(strategy)

search = ezra_engine.search
related_keywords = ezra_engine.related_keywords


def create_app():
    app = Flask(__name__)

    @app.route("/")
    def index():
        return render_template("index.html", version=get_version())

    @app.route("/search")
    def search_page():
        keyword = request.args.get("q", "")
        if not keyword:
            return redirect("/")
        else:
            return render_template(
                "search.html", version=get_version(), keyword=keyword
            )

    @app.route("/api")
    def api():
        return redirect("/static/api/index.html")

    @app.route("/api/search")
    def search():
        query = request.args.get("q")
        book = request.args.get("book", "").lower()
        top_k = request.args.get("top", 10, type=int)
        invalid_args_msg = validate_search_args(query, book, top_k)
        if invalid_args_msg:
            return make_fail_response(invalid_args_msg)

        results = ezra_engine.search(query, zh_cn=False, in_book=book, top_k=top_k)
        return {"status": "success", "data": [match.to_dict() for match in results]}

    @app.route("/api/related-keywords")
    def related_keywords():
        keyword = request.args.get("q")
        return {"status": "success", "data": ezra_engine.related_keywords(keyword)}

    def get_version():
        try:
            with open("source-context.json") as f:
                version = json.load(f)["git"]["revisionId"][:7]
        except FileNotFoundError:
            version = ""
        return version

    def validate_search_args(q, book, top):
        invalid_args_msg = {}
        if not q:
            invalid_args_msg["q"] = "Missing argument"
        if book and book not in abbr.all and book not in abbr.ranges:
            all_options = list(abbr.ranges) + abbr.all
            invalid_args_msg["book"] = f"Must be one of the following: {all_options}"
        if not 1 <= top <= 100:
            invalid_args_msg["top"] = "Must be within 1 and 100"
        return invalid_args_msg

    def make_fail_response(data):
        return {"status": "fail", "data": data}, 400

    return app



from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np

from .lang import to_simplified


class Match:
    def __init__(self, index: int, kw_scores: List[Tuple[str, float]]):
        self.index = index
        self.kw_scores = kw_scores
        self.verse = ""
        self.ref = None

    def verse_hightlight(self) -> str:
        return self._highlight_occurrences(self.verse)

    def score(self) -> float:
        return sum(sc for _, sc in self.kw_scores)

    def to_dict(self) -> dict:
        return {
            "text": self.verse,
            "ref": self.ref,
            "score": self.score(),
            "kw_scores": dict(self.kw_scores),
        }

    def _highlight_occurrences(self, text: str) -> str:
        for i, kw in enumerate(self.kw_scores):
            text = text.replace(kw[0], self._highlight(f"{kw[0]}({kw[1]:.2f})", i))
        return text

    def _highlight(self, text: str, color_code: int) -> str:
        return f"\x1b[6;30;4{color_code + 1}m{text}\x1b[0m"


class BibleSearchStrategy(ABC):
    @abstractmethod
    def search(self, keyword: str, top_k: int, range: np.ndarray = None) -> List[Match]:
        return NotImplemented

    @abstractmethod
    def related_keywords(self, keyword: str, top_k: int) -> List[str]:
        return NotImplemented


class BibleSearchEngine:
    def __init__(self, strategy: BibleSearchStrategy):
        self.strategy = strategy

    def search(
        self,
        keyword: str,
        zh_cn: bool = False,
        in_book: str = None,
        top_k: int = 10,
        verbose: bool = False,
    ) -> List[Match]:
        """
        Search for verses that match the keyword
        """
        from .resources.db import bible

        exact_idx = bible.exact_match(keyword)
        if not in_book:
            search_range = (
                np.setdiff1d(bible.idx, exact_idx) if len(exact_idx) > 0 else None
            )
        else:
            book_range = bible.book_idx(in_book)
            exact_idx = np.intersect1d(exact_idx, book_range)
            search_range = np.setdiff1d(book_range, exact_idx)
        kw_scores_exact = [(kw, 1.0) for kw in keyword.split()]
        exact_matches = [Match(index, kw_scores_exact) for index in exact_idx]
        if len(exact_idx) < top_k:
            fuzzy_matches = self.strategy.search(
                keyword, top_k - len(exact_idx), search_range
            )
            results = exact_matches + fuzzy_matches
        else:
            results = exact_matches[:top_k]

        for match in results:
            match.ref, bible_text = bible.get_record(match.index)
            if zh_cn:
                match.verse = to_simplified(bible_text)
                match.kw_scores = [
                    (to_simplified(kw), score) for kw, score in match.kw_scores
                ]
            else:
                match.verse = bible_text

        if verbose:
            print(f"Searches for {keyword}:")
            for match in results:
                print(f"Score: {match.score():.2f} {match.verse_hightlight()}")
            print()
        return results

    def related_keywords(self, keyword: str, top_k: int = 5) -> List[str]:
        return self.strategy.related_keywords(keyword, top_k)



all = [
    "gen",
    "ex",
    "lev",
    "num",
    "deut",
    "josh",
    "judg",
    "ruth",
    "1sam",
    "2sam",
    "1kin",
    "2kin",
    "1chr",
    "2chr",
    "ezra",
    "neh",
    "esth",
    "job",
    "ps",
    "prov",
    "eccl",
    "song",
    "is",
    "jer",
    "lam",
    "ezek",
    "dan",
    "hos",
    "joel",
    "amos",
    "obad",
    "jon",
    "mic",
    "nah",
    "hab",
    "zeph",
    "hag",
    "zech",
    "mal",
    "matt",
    "mark",
    "luke",
    "john",
    "acts",
    "rom",
    "1cor",
    "2cor",
    "gal",
    "eph",
    "phil",
    "col",
    "1thess",
    "2thess",
    "1tim",
    "2tim",
    "titus",
    "philem",
    "heb",
    "james",
    "1pet",
    "2pet",
    "1john",
    "2john",
    "3john",
    "jude",
    "rev",
]


def get_books(start: str, end: str):
    return all[all.index(start) : all.index(end) + 1]


ranges = {
    "nt": get_books("matt", "rev"),
    "ot": get_books("gen", "mal"),
    "gospels": get_books("matt", "john"),
    "letters": get_books("rom", "jude"),
    "torah": get_books("gen", "deut"),
    "history": get_books("josh", "esth"),
    "poetry": get_books("job", "song"),
    "prophecy": get_books("is", "mal"),
}



from importlib import resources

import numpy as np
import pandas as pd


def in_order(nums: np.array):
    return np.array_equal(nums, range(1, nums.max() + 1))


def read_bible(bible_file) -> pd.DataFrame:
    types = {
        "index": np.int16,
        "book": str,
        "chap": np.uint8,
        "vers": np.uint8,
        "text": str,
    }
    bible = pd.read_table(
        bible_file,
        sep="#",
        header=None,
        index_col=0,
        usecols=[0, 1, 2, 3, 4],
        names=list(types.keys()),
        dtype=types,
    )
    bible["book"] = bible.book.str.replace(" ", "").str.lower().astype("category")
    bible.sort_values("index", inplace=True)
    bible.reset_index(drop=True, inplace=True)

    assert len(bible.book.unique()) == 66
    assert bible.groupby("book").chap.unique().apply(in_order).all()
    vers_by_book_chap = bible.groupby(["book", "chap"]).vers
    assert vers_by_book_chap.apply(in_order).all()
    assert vers_by_book_chap.max().sum() == len(bible) == 31103

    return bible


with resources.path(__package__, "fhl_unv.dsv") as f:
    df: pd.DataFrame = read_bible(f)
    text: pd.Series = df.text
